#include "XOPStandardHeaders.h"			// Include ANSI headers, Mac headers, IgorXOP.h, XOP.h and XOPSupport.h

#include <unordered_map>
#include <string>
#include <vector>

// Author: Thomas Braun thomas dott braun ht virtuell - zuhause dott de

#include "main.h"

namespace  {

  #define ARRAY_SIZE 1024
  // Accepts multipe arguments like printf and prints them to the history
  // Custom prefixes can be also inserted, see the DEBUGPRINT macro
  // Copies only ARRAY_SIZE-2 characters in _snprintf, because we want to have space for the terminating \0 (1) and for the CR (1)
  // Checking the return value of _snprintf is not done on purpose, we just always append a \0 to be safe
  // @param A prints only if A evaluates to true
  // @param B uses silent printing (does not mark the experiment as changed) if true
  #define PRINT_TO_HISTORY(A,B,...)                \
    if (A)                                         \
    {                                              \
      char buf[ARRAY_SIZE];                        \
      _snprintf(buf,ARRAY_SIZE-2, __VA_ARGS__);    \
      buf[ARRAY_SIZE-2] = '\0';                    \
      strcat(buf,CR_STR);                          \
      if (!B)                                      \
      {                                            \
        XOPNotice(buf);                            \
      }                                            \
      else                                         \
      {                                            \
        XOPNotice2(buf,0);                         \
      }                                            \
    }
   
  // debuggingEnabled is a global boolean variable or similar which turns debug output on/off
   
  // Convenience wrapper using silent debug print
  #define DEBUGPRINT_SILENT(...) \
    PRINT_TO_HISTORY(true,true,"DEBUG: "__VA_ARGS__)
   
  // Convenience wrapper using debug print
  #define DEBUGPRINT(...) \
    PRINT_TO_HISTORY(true,false,"DEBUG: "__VA_ARGS__)
   
  // Convenience wrapper for printing
  #define HISTPRINT(...) \
    PRINT_TO_HISTORY(true,false,__VA_ARGS__)
   
  // Convenience wrapper for silent printing
  #define HISTPRINT_SILENT(...) \
    PRINT_TO_HISTORY(true,true,__VA_ARGS__)

  // Igor aborts if it encounters an unhandled exception, therefore every operation
  // needs enclosing BEGIN_OUTER_CATCH and END_OUTER_CATCH macros
  #define BEGIN_OUTER_CATCH  try{
  #define END_OUTER_CATCH    }\
                catch(...){\
                  return 0;\
                }

  typedef std::tr1::unordered_map<std::string,waveHndl> HashMap;

  class GlobalData
  {
    public:
      GlobalData()
      {
        waveMap.max_load_factor(0.7f);
        waveMap.bucket_size(100);   
      }

      HashMap waveMap;
  };

  GlobalData storage;

  const int MIN_IGOR_VERSION            = 630;

  int RegisterFetchWave(void)
  {
    const char* cmdTemplate;
    const char* runtimeNumVarList;
    const char* runtimeStrVarList;

    // NOTE: If you change this template, you must change the FetchWaveRuntimeParams structure as well.
    cmdTemplate = "FetchWave /L=string:wvName /DEST=DataFolderAndName:{wv,wave}";
    runtimeNumVarList = "";
    runtimeStrVarList = "";
    return RegisterOperation(cmdTemplate, runtimeNumVarList, runtimeStrVarList, sizeof(FetchWaveRuntimeParams), (void*)ExecuteFetchWave, 0);
  }

  int RegisterUpdateIndex( void )
  {
    const char* cmdTemplate;
    const char* runtimeNumVarList;
    const char* runtimeStrVarList;

    // NOTE: If you change this template, you must change the UpdateIndexRuntimeParams structure as well.
    cmdTemplate = "UpdateIndex";
    runtimeNumVarList = "";
    runtimeStrVarList = "";
    return RegisterOperation(cmdTemplate, runtimeNumVarList, runtimeStrVarList, sizeof(UpdateIndexRuntimeParams), (void*)ExecuteUpdateIndex, 0);
  }

  // Register any operations with Igor
  int RegisterOperations()		
  {
    if (int result = RegisterUpdateIndex())
      return result;

    if (int result = RegisterFetchWave())
      return result;

    return 0;
  }

  /*
    Splits string separated by sepChar into a vector of strings
  */
  void splitString(const std::string& string, const char* sepChar, std::vector<std::string> &list)
  {
    if (sepChar == NULL)
    {
      return;
    }

    list.clear();
    std::string stringCopy = string;
    int pos = -1;
    int offset = 0;

    stringCopy.append(sepChar); // add ; at the end to make the list complete, double ;; are no problem

    while ((pos = stringCopy.find(sepChar, offset)) != std::string::npos)
    {
      if (pos == offset) // skip empty element
      {
        offset++;
        continue;
      }

      list.push_back(stringCopy.substr(offset, pos - offset));
      offset = pos + 1;
    }
  }

  /*
    Convert a XOP string handle to a std::string
  */
  void convertHandleToString(Handle strHandle, std::string& str)
  {
    str.clear();
    // for both cases we return an empty string
    if (strHandle == NULL || GetHandleSize(strHandle) == 0L)
    {
      return;
    }

    const int handleSize = GetHandleSize(strHandle);

    str = std::string(*strHandle, handleSize);
  }

} // anonymous namespace 

extern "C" void XOPEntry()
{	
	XOPIORecResult result = 0;
	
	//switch (GetXOPMessage())
 // {
	//	// We don't need to handle any messages for this XOP.
	//}
	SetXOPResult(result);
}

HOST_IMPORT int main(IORecHandle ioRecHandle)
{
	int result;
	
	XOPInit(ioRecHandle);
	SetXOPEntry(XOPEntry);
	
  if (igorVersion < MIN_IGOR_VERSION)
  {			
		SetXOPResult(OLD_IGOR);
		return EXIT_FAILURE;
	}

	if (result = RegisterOperations())
  {
		SetXOPResult(result);
		return EXIT_FAILURE;
	}
	
	SetXOPResult(0);
	return EXIT_SUCCESS;
}

extern "C" int ExecuteUpdateIndex( UpdateIndexRuntimeParamsPtr p )
{
  BEGIN_OUTER_CATCH

  Handle listHandle = NewHandle(0);

  if( int err = WaveList(listHandle,"*",";","") )
    return err;

  std::string listOfWavesAsString;
  convertHandleToString(listHandle,listOfWavesAsString);
  DisposeHandle(listHandle);

  typedef std::vector<std::string> StringVector;
  typedef std::vector<std::string>::iterator StringVectorIt;

  StringVector listOfWaves;
  splitString(listOfWavesAsString,";",listOfWaves);

  storage.waveMap.clear();

  for (StringVectorIt it = listOfWaves.begin(); it != listOfWaves.end(); it++)
    if (!it->empty())
      storage.waveMap[*it] = FetchWave(it->c_str());

  DEBUGPRINT("bucket count=%d, load factor=%f",storage.waveMap.bucket_count(),storage.waveMap.load_factor());

  return 0;

  END_OUTER_CATCH
}

extern "C" int ExecuteFetchWave( FetchWaveRuntimeParamsPtr p )
{
  BEGIN_OUTER_CATCH

  if (!p->LFlagEncountered || !p->DESTFlagEncountered)
  {
      return MISSING_PARAMS;
  }

  std::string wvName;
  convertHandleToString(p->wvName,wvName);

  HashMap::const_iterator it = storage.waveMap.find(wvName);

  if (it == storage.waveMap.end())
  {
    DEBUGPRINT("Could not find wave named %s",wvName.c_str());
    return OLD_INDEX;
  }

  if(int result = SetOperationWaveRef(it->second,p->DESTFlagParamsSet[0]))
    return result;

  return 0;

  END_OUTER_CATCH
}
